import re
from datetime import datetime
import pdfplumber
import pandas as pd

# from pdfminer.high_level import extract_text
import tabula


class StatementParser:
    def __init__(self, file_path):
        self.file_path = file_path

    def extract_text(self):
        with pdfplumber.open(self.file_path) as pdf:
            full_text = "\n".join(page.extract_text() for page in pdf.pages)
        return full_text

    def extract_table(self):
        tables = tabula.read_pdf(self.file_path, pages="all")
        return tables

    def extract_transactions(self):
        # This method should be implemented by each subclass
        raise NotImplementedError


class DBSCreditCardStatementParser(StatementParser):
    def extract_card_last_4_digits(self):
        full_text = self.extract_text()
        card_number_pattern = r"(\d{4}-\d{4}-\d{4}-\d{4})"
        card_number_match = re.search(card_number_pattern, full_text)
        if card_number_match:
            card_number = card_number_match.group()
            last_4_digits = card_number.split("-")[-1]
        else:
            raise ValueError("Card number not found in the statement.")
        return last_4_digits

    def extract_previous_balance(self):
        full_text = self.extract_text()
        prev_balance_pattern = r"PREVIOUS BALANCE ([0-9,]+\.\d{2})( CR)?"
        prev_balance_match = re.search(prev_balance_pattern, full_text)
        if prev_balance_match:
            prev_balance = self.clean_amount(prev_balance_match.group(1))
        else:
            raise ValueError("Previous balance not found in the statement.")
        return prev_balance

    def extract_transactions(self):
        full_text = self.extract_text()
        pattern = r"(\d{2} \w{3})\s(\d{2} \w{3})\s(.*?)\s([0-9,]+\.\d{2}(?: CR)?)"
        matches = re.findall(pattern, full_text, re.DOTALL)
        df = pd.DataFrame(
            matches, columns=["TRANS DATE", "POST DATE", "DESCRIPTION", "AMOUNT HKD"]
        )
        df["TRANS DATE"] = df["TRANS DATE"].apply(lambda x: self.add_year_to_date(x))
        df["POST DATE"] = df["POST DATE"].apply(lambda x: self.add_year_to_date(x))
        df["AMOUNT HKD"] = df["AMOUNT HKD"].apply(self.clean_amount)
        df.insert(0, "Card Last 4 Digits", self.extract_card_last_4_digits())
        df["RUNNING BALANCE"] = (
            df["AMOUNT HKD"].cumsum() + self.extract_previous_balance()
        )
        return df

    def clean_amount(self, amount):
        if "CR" in amount:
            amount = -float(amount.replace(",", "").replace("CR", ""))
        else:
            amount = float(amount.replace(",", ""))
        return amount

    def extract_statement_date(self):
        full_text = self.extract_text()
        statement_date_pattern = r"STATEMENT DATE (\d{2} \w{3} \d{4})"
        statement_date_match = re.search(statement_date_pattern, full_text)
        if statement_date_match:
            statement_date = datetime.strptime(
                statement_date_match.group(1), "%d %b %Y"
            )
        else:
            raise ValueError("Statement date not found in the statement.")
        return statement_date

    def add_year_to_date(self, date):
        statement_date = self.extract_statement_date()
        date_with_year = datetime.strptime(
            date + " " + str(statement_date.year), "%d %b %Y"
        )
        if date_with_year > statement_date:
            date_with_year = datetime.strptime(
                date + " " + str(statement_date.year - 1), "%d %b %Y"
            )
        return date_with_year


class BankSavingAccountStatementParser(StatementParser):
    def extract_transactions(self):
        # Code specific to Bank Saving Account statements
        pass


class FundAccountStatementParser(StatementParser):
    def extract_transactions(self):
        # Code specific to Fund Account statements
        pass
